import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:meta/meta.dart';

import 'persistence.dart';

typedef AsyncFileGetter = Future<File> Function();

class JsonFilePersistence<State> extends Persistence<State> {
  const JsonFilePersistence({
    @required this.file,
    @required this.converter,
  });

  final JsonConverter<State> converter;
  final AsyncFileGetter file;

  File get backupFile => File('${file}.backup');

  @override
  Future<void> clear() => file().then((file) => file.delete());

  @override
  Future<State> read() async {
    final file = await this.file();

    if (file.existsSync()) {
      /// Veryfing that hashcode is equivalent to avoid corrupted content
      final content = await file.readAsString();
      final splits = content.indexOf(';');
      int hashCode;
      String json;

      if (splits >= 0) {
        hashCode = int.parse(content.substring(0, splits));
        json = content.substring(splits);
      }

      /// File is corrupted, trying to load backup file
      if (json == null || hashCode != json.hashCode) {
        json = null;

        final content = await backupFile.readAsString();
        final splits = content.indexOf(';');

        if (splits >= 0) {
          hashCode = int.parse(content.substring(0, splits));
          json = content.substring(splits);
        }

        /// Even backup file is corrupted
        if (json == null || hashCode != json.hashCode) {
          return null;
        }
      }

      return converter.fromJson(jsonDecode(json));
    }

    return null;
  }

  @override
  Future<void> write(State state) async {
    final file = await this.file();
    final contents = jsonEncode(converter.toJson(state));
    await file.writeAsString('${contents.hashCode};' + contents);

    /// We create a backup in case a following write is corrupted
    await file.copy(backupFile.path);
  }
}

abstract class BaseJsonConverter<State> {
  const BaseJsonConverter();
  State fromJson(dynamic value);
  dynamic toJson(State state);
}

class JsonConverter<State> extends BaseJsonConverter<State> {
  const JsonConverter({
    @required FromJsonConverter<State> fromJson,
    @required ToJsonConverter<State> toJson,
  })  : assert(fromJson != null),
        assert(toJson != null),
        _fromJson = fromJson,
        _toJson = toJson;
  final FromJsonConverter<State> _fromJson;
  final ToJsonConverter<State> _toJson;

  @override
  State fromJson(dynamic value) => _fromJson(value);

  @override
  dynamic toJson(State state) => _toJson(state);
}

typedef FromJsonConverter<State> = State Function(dynamic value);

typedef ToJsonConverter<State> = dynamic Function(State state);
